from __future__ import annotations

from typing import TYPE_CHECKING, List, Union

if TYPE_CHECKING:
    from pathlib import Path


import spikeinterface.extractors as se
from spikeinterface import append_recordings

from ..data_classes.preprocessing import PreprocessingData
from ..data_classes.sorting import SortingData


def load_spikeglx_data(
    base_path: Union[Path, str], sub_name: str, run_names: Union[List[str], str]
) -> PreprocessingData:
    """
    Load raw SpikeGLX data (in rawdata). If multiple runs are selected
    in run_names, these will be stored as segments on a SpikeInterface
    recording object.

    Parameters
    -----------

    base_path : Union[Path, str]
        Path where the rawdata folder containing subjects.

    sub_name : str
        Subject to preprocess. The subject top level dir should reside in
        base_path/rawdata/ .

    run_names : Union[List[str], str],
        The SpikeGLX run name (i.e. not including the gate index). This can
        also be a list of run names.

    Returns
    -------

    PreprocessingData class containing SpikeInterface recording object and information
    on the data filepaths.
    """
    preprocess_data = PreprocessingData(base_path, sub_name, run_names)

    all_recordings = []
    all_sync = []
    for run_path in preprocess_data.all_run_paths:
        with_sync, without_sync = [
            se.read_spikeglx(
                folder_path=run_path,
                stream_id="imec0.ap",
                all_annotations=True,
                load_sync_channel=sync,
            )
            for sync in [True, False]
        ]
        all_recordings.append(without_sync)
        sync_channel_id = with_sync.get_channel_ids()[-1]
        all_sync.append(with_sync.channel_slice(channel_ids=[sync_channel_id]))

    preprocess_data["0-raw"] = append_recordings(all_recordings)
    preprocess_data.sync = append_recordings(all_sync)

    return preprocess_data


def load_data_for_sorting(
    preprocessed_data_path: Path,
    concatenate: bool = True,
) -> SortingData:
    """
    Returns the previously preprocessed PreprocessingData and
    recording object loaded from the preprocess path.

    During sorting, preprocessed data is saved to
    derivatives/<sub level dirs>/preprocessed. The spikeinterface
    recording (si_recording) and PreprocessingData (data_class.pkl) are saved.

    Parameters
    ----------
    preprocessed_data_path : Path
        Path to the preprocessed folder, containing the binary si_recording
        of the preprocessed data and the data_class.pkl containing all filepath
        information.

    concatenate : bool
        If True, the multi-segment recording object will be concatenated
        together. This is used prior to sorting. Segments should be
        experimental runs.

    Returns
    -------
    sorting_data : SortingData
        The sorting_data dict with the loaded spikeinterface
        recording attached to the '0-preprocessed' field.
    """
    sorting_data = SortingData(
        preprocessed_data_path,
    )

    sorting_data.load_preprocessed_binary(concatenate)

    return sorting_data



from collections import UserDict
from collections.abc import ItemsView, KeysView, ValuesView


class BaseUserDict(UserDict):
    """
    Base UserDict that implements the
    keys(), values() and items() convenience functions.
    """

    def __init__(self):
        super(BaseUserDict, self).__init__()

    def keys(self) -> KeysView:
        return self.data.keys()

    def items(self) -> ItemsView:
        return self.data.items()

    def values(self) -> ValuesView:
        return self.data.values()



from pathlib import Path

from swc_ephys.pipeline.full_pipeline import run_full_pipeline

base_path = Path(
    "/ceph/neuroinformatics/neuroinformatics/scratch/jziminski/ephys/test_data/steve_multi_run/1119617/time-short"
)
#    r"C:\data\ephys\test_data\steve_multi_run\1119617\time-short"
sub_name = "1119617"
run_names = [
    "1119617_LSE1_shank12",
    "1119617_posttest1_shank12",
    "1119617_pretest1_shank12",
]

config_name = "default"
sorter = "mountainsort5"  #  "kilosort2_5"  # "spykingcircus" # mountainsort5

if __name__ == "__main__":
    run_full_pipeline(
        base_path,
        sub_name,
        run_names,
        config_name,
        sorter,
        existing_preprocessed_data="load_if_exists",
        existing_sorting_output="overwrite",
        overwrite_postprocessing=True,
        slurm_batch=False,
    )



from swc_ephys.pipeline.load_data import load_spikeglx_data
from swc_ephys.pipeline.preprocess import preprocess
from swc_ephys.pipeline.visualise import visualise

base_path = r"X:\neuroinformatics\scratch\jziminski\ephys\test_data\steve_multi_run\1119617\time-short"
sub_name = "1119617"
run_names = "1119617_LSE1_shank12"

data = load_spikeglx_data(base_path, sub_name, run_names)

data = preprocess(data, pp_steps="default")

visualise(
    data,
    steps=["all"],
    mode="map",
    as_subplot=True,
    show_channel_ids=True,
    time_range=(0, 1),
    run_number=1,
)



import json
from typing import Dict, List, Tuple, Union

import numpy as np
import spikeinterface.preprocessing as spre

from ..configs import configs
from ..data_classes.preprocessing import PreprocessingData
from ..utils import utils


def preprocess(
    preprocess_data: PreprocessingData,
    pp_steps: Union[Dict, str],
    verbose: bool = True,
) -> PreprocessingData:
    """
    Returns an updated PreprocessingData dictionary of SpikeInterface
    recording objects setup in the order and with the options specified
    in pp_steps.

    Spikeinterface preprocessing is lazy - no preprocessing is done
    until the data is written to file or the get_traces() method on
    the recording object is called.

    Parameters
    ----------
    preprocess_data : PreprocessingData
        A preprocessing data object that has as attributes the
        paths to rawdata. The pp_steps attribute is set on
        this class during execution of this function.

    pp_steps: either a pp_steps dictionary, or name of valid
              preprocessing .yaml file (without hte yaml extension).
              See configs/configs.py for details.

    verbose : bool
        If True, messages will be printed to console updating on the
        progress of preprocessing / sorting.

    Returns
    -------
    preprocess_data : PreprocessingData
        Preprocessing data class with pp_steps updated and
        dictionary field filled with key-value pairs indicating
        the name and order of preprocessing keys, and value containing
        associated SpikeInterface recording objects.

    """
    if isinstance(pp_steps, str):
        pp_steps_to_run, _, _ = configs.get_configs(pp_steps)
    else:
        pp_steps_to_run = pp_steps

    pp_funcs = get_pp_funcs()

    checked_pp_steps, pp_step_names = check_and_sort_pp_steps(pp_steps_to_run, pp_funcs)

    preprocess_data.set_pp_steps(pp_steps_to_run)

    for step_num, pp_info in checked_pp_steps.items():
        perform_preprocessing_step(
            step_num, pp_info, preprocess_data, pp_step_names, pp_funcs, verbose
        )

    return preprocess_data


def check_and_sort_pp_steps(pp_steps: Dict, pp_funcs: Dict) -> Tuple[Dict, List[str]]:
    """
    Sort the preprocessing steps dictionary by order to be run (based on the
    keys) and check the dictionary is valid.

    Parameters
    ----------
    pp_steps : Dict
        A dictionary with keys as numbers indicating the order that
        preprocessing steps are run (starting at "1"). The values are a
        (preprocessing name, preprocessing kwargs) tuple containing the
        spikeinterface preprocessing function name, and kwargs to pass to it.

    pp_funcs : Dict
        A dictionary linking preprocessing step names to the underlying
        SpikeInterface function objects that conduct the preprocessing.

    Returns
    -------
    pp_steps :Dict
        The checked preprocessing steps dictionary.

    pp_step_names : List
        Preprocessing step names (e.g. "bandpass_filter") in order
        that they are to be run.

    Notes
    -------
    This will soon be deprecated and replaced by validation
    of the config file itself on load.
    """
    validate_pp_steps(pp_steps)
    pp_step_names = [item[0] for item in pp_steps.values()]

    # Check the preprocessing function names are valid and print steps used
    canonical_step_names = list(pp_funcs.keys())

    for user_passed_name in pp_step_names:
        assert (
            user_passed_name in canonical_step_names
        ), f"{user_passed_name} not in allowed names: ({canonical_step_names}"

    utils.message_user(
        f"\nThe preprocessing options dictionary is "
        f"{json.dumps(pp_steps, indent=4, sort_keys=True)}"
    )

    return pp_steps, pp_step_names


def validate_pp_steps(pp_steps: Dict):
    """
    Ensure the pp_steps dictionary of preprocessing steps to
    has number-order that makes sense. The preprocessing step numbers
    should start 1 at, and increase by 1 for each subsequent step.
    """
    assert all(
        key.isdigit() for key in pp_steps.keys()
    ), "pp_steps keys must be integers"

    key_nums = [int(key) for key in pp_steps.keys()]

    assert np.min(key_nums) == 1, "dict keys must start at 1"

    if len(key_nums) > 1:
        diffs = np.diff(key_nums)
        assert np.unique(diffs).size == 1, "all dict keys must increase in steps of 1"
        assert diffs[0] == 1, "all dict keys must increase in steps of 1"


def perform_preprocessing_step(
    step_num: str,
    pp_info: Tuple[str, Dict],
    preprocess_data: PreprocessingData,
    pp_step_names: List[str],
    pp_funcs: Dict,
    verbose: bool = True,
) -> None:
    """
    Given the preprocessing step and preprocess_data UserDict containing
    spikeinterface BaseRecordings, apply a preprocessing step to the
    last preprocessed recording and save the recording object to PreprocessingData.
    For example, if step_num = "3", get the recording of the second
    preprocessing step from preprocess_data and apply the 3rd preprocessing step
    as specified in pp_info.

    Parameters
    ----------
    step_num : str
        Preprocessing step to run (e.g. "1", corresponds to the key in pp_dict).

    pp_info : Tuple[str, Dict]
        Preprocessing name, preprocessing kwargs) tuple (the value from
        the pp_dict).

    preprocess_data : PreprocessingData
        swc_ephys PreprocessingData class (a UserDict in which key-values are
        the preprocessing chain name : spikeinterface recording objects).

    pp_step_names : List[str]
        Ordered list of preprocessing step names that are being
        applied across the entire preprocessing chain.

    pp_funcs : Dict
        The canonical SpikeInterface preprocessing functions. The key
        are the function name and value the function object.

    verbose : bool
        If True, messages will be printed to console updating on the
        progress of preprocessing / sorting.
    """
    pp_name, pp_options = pp_info

    utils.message_user(
        f"Running preprocessing step: {pp_name} with options {pp_options}", verbose
    )

    last_pp_step_output, __ = utils.get_dict_value_from_step_num(
        preprocess_data, step_num=str(int(step_num) - 1)
    )

    new_name = f"{step_num}-" + "-".join(["raw"] + pp_step_names[: int(step_num)])

    confidence_check_pp_func_name(pp_name, pp_funcs)

    if isinstance(last_pp_step_output, Dict):
        preprocess_data[new_name] = {
            k: pp_funcs[pp_name](v, **pp_options)
            for k, v in last_pp_step_output.items()
        }
    else:
        preprocess_data[new_name] = pp_funcs[pp_name](last_pp_step_output, **pp_options)


def confidence_check_pp_func_name(pp_name, pp_funcs):
    """
    Ensure that the correct preprocessing function is retrieved. This
    essentially checks the get_pp_funcs dictionary is correct.

    TODO
    ----
    This should be a standalone test, not incorporated into the package.
    """
    func_name_to_class_name = "".join([word.lower() for word in pp_name.split("_")])

    if pp_name == "silence_periods":
        assert pp_funcs[pp_name].__name__ == "SilencedPeriodsRecording"  # TODO: open PR
    elif isinstance(pp_funcs[pp_name], type):
        assert (
            func_name_to_class_name in pp_funcs[pp_name].__name__.lower()
        ), "something is wrong in func dict"

    else:
        assert pp_funcs[pp_name].__name__ == pp_name


def remove_channels(recording, bad_channel_ids):
    return recording.remove_channels(bad_channel_ids)


def get_pp_funcs() -> Dict:
    """
    Get the spikeinterface preprocessing function
    from its name.

    TODO
    -----
    It should be possible to generate this on the fly from
    SI __init__ rather than hard code like this
    """
    pp_funcs = {
        "phase_shift": spre.phase_shift,
        "bandpass_filter": spre.bandpass_filter,
        "common_reference": spre.common_reference,
        "blank_saturation": spre.blank_staturation,
        "center": spre.center,
        "clip": spre.clip,
        "correct_lsb": spre.correct_lsb,
        "correct_motion": spre.correct_motion,
        "depth_order": spre.depth_order,
        "filter": spre.filter,
        "gaussian_bandpass_filter": spre.gaussian_bandpass_filter,
        "highpass_filter": spre.highpass_filter,
        "interpolate_bad_channels": spre.interpolate_bad_channels,
        "normalize_by_quantile": spre.normalize_by_quantile,
        "notch_filter": spre.notch_filter,
        "remove_artifacts": spre.remove_artifacts,
        "remove_channels": remove_channels,
        "resample": spre.resample,
        "scale": spre.scale,
        "silence_periods": spre.silence_periods,
        "whiten": spre.whiten,
        "zscore": spre.zscore,
    }

    return pp_funcs



from __future__ import annotations

import time
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Dict, Literal, Optional, Union

import matplotlib.pyplot as plt
import numpy
import numpy as np
import pandas as pd
import spikeinterface as si
from spikeinterface import curation
from spikeinterface.extractors import NpzSortingExtractor

from ..configs.configs import get_configs
from ..data_classes.sorting import SortingData
from ..pipeline.load_data import load_data_for_sorting
from ..utils import utils
from ..utils.custom_types import HandleExisting
from .waveform_compare import get_waveform_similarity

if TYPE_CHECKING:
    from spikeinterface import WaveformExtractor
    from spikeinterface.core import BaseSorting

MATRIX_BACKEND: Literal["numpy", "jax"]
try:
    MATRIX_BACKEND = "jax"
except ImportError:
    warnings.warn(
        "The module Jax was not found. Waveform similarity matrix calculation"
        "will be very slow. Use `pip install jax jaxlib` to install."
    )
    MATRIX_BACKEND = "numpy"


# --------------------------------------------------------------------------------------
# Run Postprocessing
# --------------------------------------------------------------------------------------


def run_postprocess(
    sorting_data: Union[Path, str, SortingData],
    sorter: str,
    existing_waveform_data: HandleExisting = "load_if_exists",
    postprocessing_to_run: Union[Literal["all"], Dict] = "all",
    verbose: bool = True,
    waveform_options: Optional[Dict] = None,
) -> None:
    """
    Run post-processing, including ave quality metrics on sorting
    output to a quality_metrics.csv file.

    Parameters
    ----------

    sorting_data : Union[Path, str, SortingData]
        The path to the 'preprocessed' folder in the subject / run
        folder used for sorting or a SortingData object. If a

        SortingData object, the path will be read from the
        `preprocessed_data_path` attribute.

    sorter : str
        The name of the sorter (e.g. "kilosort2_5").

    existing_waveform_data : Literal["overwrite", "load_if_exists", "fail_if_exists"]
        Determines how existing preprocessed data (e.g. from a prior pipeline run)
        is treated.
            "overwrite" : will overwrite any existing preprocessed data output. This will
                          delete the 'preprocessed' folder. Therefore, never save
                          derivative work there.
            "load_if_exists" : will search for existing data and load if it exists.
                               Otherwise, will use the preprocessing from the
                               current run.
            "fail_if_exists" : If existing preprocessed data is found, an error
                               will be raised.

    postprocessing_to_run : Union[Literal["all"], Dict]
        Specify the postprocessing to run. By default, "all" will run
        all available postprocessing. Otherwise, provide a dict of
        including postprocessing to run e.g. {"quality_metrics: True"}.

    verbose : bool
        If True, messages will be printed to console updating on the
        progress of preprocessing / sorting.

    waveform_options: Dict
        A dictionary containing options passed to SpikeInterface's
        `extract_waveforms()` function as kwargs.
    """
    if not isinstance(sorting_data, SortingData):
        sorting_data = load_data_for_sorting(
            Path(sorting_data),
        )
    assert isinstance(sorting_data, SortingData), "type narrow `sorting_data`."

    # Create / load waveforms
    if waveform_options is None:
        _, _, waveform_options = get_configs("default")

    sorting_data.set_sorter_output_paths(sorter)

    utils.message_user(
        f"Quality Checks: sorting path used: {sorting_data.sorter_run_output_path}",
        verbose,
    )

    waveforms = run_or_get_waveforms(
        sorting_data, existing_waveform_data, waveform_options, sorter, verbose
    )

    # Perform postprocessing
    run_settings = handle_postprocessing_to_run(postprocessing_to_run)

    if run_option(run_settings, "quality_metrics"):
        save_quality_matrics(waveforms, sorting_data)

    if run_option(run_settings, "unit_locations"):
        save_unit_locations(waveforms, sorting_data)

    if run_option(run_settings, "template_plots"):
        save_plots_of_templates(sorting_data.postprocessing_output_path, waveforms)

    if run_option(run_settings, "waveform_similarity"):
        save_waveform_similarities(
            sorting_data.postprocessing_output_path, waveforms, MATRIX_BACKEND
        )


# Sorting Loader -----------------------------------------------------------------------


def run_option(run_settings: Dict, option: str):
    return option in run_settings and run_settings[option]


def run_or_get_waveforms(
    sorting_data: SortingData,
    existing_waveform_data: HandleExisting,
    waveform_options: Dict,
    sorter: str,
    verbose: bool,
):
    """
    How to handle existing waveform output, either load, fail if exists or
    overwrite.
    """
    if (
        sorting_data.postprocessing_output_path.is_dir()
        and existing_waveform_data == "load_if_exists"
    ):
        utils.message_user(
            f"Loading existing waveforms from: "
            f"{sorting_data.postprocessing_output_path}",
            verbose,
        )
        waveforms = si.load_waveforms(sorting_data.postprocessing_output_path)

    elif (
        sorting_data.postprocessing_output_path.is_dir()
        and existing_waveform_data == "fail_if_exists"
    ):
        raise RuntimeError(
            f"Waveforms exist at {sorting_data.postprocessing_output_path} but "
            f"`existing_waveform_data` is 'fail_if_exists'."
        )
    else:
        utils.message_user(
            f"Saving waveforms to {sorting_data.postprocessing_output_path}"
        )

        sorting_without_excess_spikes = load_sorting_output(sorting_data, sorter)

        waveforms = si.extract_waveforms(
            sorting_data.data["0-preprocessed"],
            sorting_without_excess_spikes,
            folder=sorting_data.postprocessing_output_path,
            use_relative_path=True,
            overwrite=True,
            **waveform_options,
        )

    return waveforms


def handle_postprocessing_to_run(postprocessing_to_run: Union[Literal["all"], Dict]):
    """
    Set to run all postprocessing steps. If user-dict is provided,
    ensure it contains expected keys / values.
    """
    run_settings = {
        "quality_metrics": True,
        "unit_locations": True,
        "template_plots": True,
        "waveform_similarity": True,
    }

    if postprocessing_to_run == "all":
        return run_settings
    else:
        assert isinstance(postprocessing_to_run, Dict)
        assert all(
            [key in run_settings.keys() for key in postprocessing_to_run.keys()]
        ), (
            f"At least one option in `postprocessing_to_run` is invalid. Must be"
            f"one of {run_settings.keys()}"
        )
        assert all(
            [isinstance(value, bool) for value in postprocessing_to_run.values()]
        ), "`postprocessing_to_run` values must be `True` or `False`."

        run_settings = postprocessing_to_run

        return run_settings


def load_sorting_output(sorting_data: SortingData, sorter: str) -> BaseSorting:
    """
    Load the output of a sorting run as a SpikeInterface SortingExtractor
    object.

    It is assumed sorting is performed on concatenated runs. As such,
    only a single sorting segment is expected, and error raised
    if there are more.

    Automatically remove empty units, which complicate downstream
    processing. Also, remove excess spikes, which are caused by
    strange behaviour of KS returning spike times that occur
    outside the number of samples in a recording. This is required
    for waveform extraction in Spikeinterface.

    Parameters
    ----------

    sorting_data : SortingData
        An swc_ephys SortingData object holding information about sorting

    sorter : str
        The sorter used (e.g. "kilosort2_5")

    TODO
    ----
    In newer SpikeInterface version, `remove_empty_units()` is automatically
    applied during `KiloSortSortingExtractor()`. We can simply use
    this default rather than applying again here after pinning to newer
    SI version. See  https://github.com/SpikeInterface/spikeinterface/issues/1760
    """
    if not sorting_data.sorter_run_output_path.is_dir():
        raise FileNotFoundError(
            f"{sorter} output was not found at "
            f"{sorting_data.sorter_run_output_path}.\n"
            f"Quality metrics will not be generated."
        )

    assert (
        len(sorting_data) == 1
    ), "unexpected number of entries in `sorting_data` dict."

    recording = sorting_data[sorting_data.init_data_key]

    if "kilosort" in sorter:
        sorting = si.extractors.read_kilosort(
            folder_path=sorting_data.sorter_run_output_path,
            keep_good_only=False,
        )
    elif sorter == "mountainsort5":
        sorting = NpzSortingExtractor(
            (sorting_data.sorter_run_output_path / "firings.npz").as_posix()
        )

    elif sorter == "tridesclous":
        sorting = si.extractors.read_tridesclous(
            folder_path=sorting_data.sorter_run_output_path.as_posix()
        )

    elif sorter == "spykingcircus":
        sorting = si.extractors.read_spykingcircus(
            folder_path=sorting_data.sorter_run_output_path.as_posix()
        )

    sorting = sorting.remove_empty_units()
    sorting_without_excess_spikes = curation.remove_excess_spikes(sorting, recording)

    return sorting_without_excess_spikes


# Helpers ------------------------------------------------------------------------------


def save_quality_matrics(waveforms: WaveformExtractor, sorting_data: SortingData):
    """ """
    quality_metrics = si.qualitymetrics.compute_quality_metrics(waveforms)
    quality_metrics.to_csv(sorting_data.quality_metrics_path)
    utils.message_user(f"Quality metrics saved to {sorting_data.quality_metrics_path}")


def save_unit_locations(waveforms: WaveformExtractor, sorting_data: SortingData):
    """ """
    unit_locations = si.postprocessing.compute_unit_locations(
        waveforms, outputs="by_unit"
    )
    unit_locations_pandas = pd.DataFrame.from_dict(
        unit_locations, orient="index", columns=["x", "y"]
    )
    unit_locations_pandas.to_csv(sorting_data.unit_locations_path)

    utils.message_user(f"Unit locations saved to {sorting_data.unit_locations_path}")


def save_waveform_similarities(
    postprocessing_output_path: Path,
    waveforms: WaveformExtractor,
    backend: Literal["jax", "numpy"],
):
    """
    Save waveform similarity matrices as a csv file. This is a
    num_waveform x num_waveform matrix holding the cosine similarity
    between all waveforms for a unit.

    The .csv file will contain header and index for the spike time of
    each row.

    Parameters
    ---------

    postprocessing_output_path : Path
        Pathlib object holding the output path where waveforms are saved
        and /images will be written.

    waveforms : WaveformExtractor
        Spikeinterface WaveformExtractor object.

    backend : Literal["jax", "numpy"]
        The backend with which to run waveform comparisons. Jax is around 50
        times faster in this case.
    """
    utils.message_user("Saving waveform similarity matrices...\n")

    matrices_out_path = postprocessing_output_path / "similarity_matrices"
    matrices_out_path.mkdir(exist_ok=True)

    t = time.perf_counter()
    for unit_id in waveforms.sorting.get_unit_ids():
        sim_matrix, spike_times = get_waveform_similarity(waveforms, unit_id, backend)
        sim_matrix_pd = pd.DataFrame(sim_matrix, columns=spike_times, index=spike_times)
        sim_matrix_pd.to_csv(
            matrices_out_path / f"waveform_similarity_unit_{unit_id}.csv"
        )

    utils.message_user(f"Waveform similarity matrices saved to: {matrices_out_path}")
    utils.message_user(
        f"Saving waveform similarity matrices took: {time.perf_counter() - t}"
    )


# --------------------------------------------------------------------------------------
# Save imagines
# --------------------------------------------------------------------------------------


def save_plots_of_templates(
    postprocessing_output_path: Path, waveforms: WaveformExtractor
):
    """
    Save a plot of all templates in 'waveforms/images' folder. The plot
    displays the template waveform are calculated in two ways
        1) as the mean over channels (the channels here are defined by the
           sparsity settings during waveform extraction).
        2) as the channel in which the waveform signal is strongest (as
           determined by the channel minimum value).
           TODO: this is  mostly experimental for testing (for visualisation only).

    The purpose of these plots is to provide a quick overview of the
    extract templates. Proper investigations on unit quality should
    be done in Phy.

    Parameters
    ------

    postprocessing_output_path : Path
        Pathlib object holding the output path where waveforms are saved
        and /images will be written.

    waveforms : WaveformExtractor
        Spikeinterface WaveformExtractor object.

    """
    t = time.perf_counter()
    utils.message_user("Saving template images...\n")

    fs = waveforms.sampling_frequency
    n_samples = waveforms.nsamples
    time_ = np.arange(n_samples) / fs * 1000

    y_label = "Voltage (uV)" if waveforms.return_scaled else "Voltage (unscaled)"

    for idx, unit_id in enumerate(waveforms.sorting.get_unit_ids()):
        unit_template = waveforms.get_template(unit_id)
        plt.plot(time_, np.mean(unit_template, axis=1))

        peak_chan_idx = np.argmin(np.min(unit_template, axis=0))
        plt.plot(time_, unit_template[:, peak_chan_idx])

        plt.legend(["mean across channels", "max signal channel"])
        plt.xlabel("Time (ms)")
        plt.ylabel(y_label)
        plt.title(f"Unit {unit_id} Template")

        output_folder = postprocessing_output_path / "template_plots"
        output_folder.mkdir(exist_ok=True)
        plt.savefig(
            postprocessing_output_path / "template_plots" / f"unit_{unit_id}.png"
        )
        plt.clf()

    utils.message_user(f"Saving plots of templates took: {time.perf_counter() - t}")



import glob
import os
from pathlib import Path
from typing import Dict, Tuple

import yaml

from ..utils import utils


def get_configs(name: str) -> Tuple[Dict, Dict, Dict]:
    """
    Loads the config yaml file in the same folder
    (swc_ephys/configs) containing preprocessing (pp)
    and sorter options.

    Once loaded, the list containing preprocesser name
    and kwargs is cast to tuple. This keeps the type
    checker happy while not requiring a tuple
    in the .yaml which require ugly tags.

    Parameters
    ----------

    name: name of the configs to load. Should not include the
          .yaml suffix.

    Returns
    -------

    pp_steps : a dictionary containing the preprocessing
               step order (keys) and a [pp_name, kwargs]
               list containing the spikeinterface preprocessing
               step and keyword options.

    sorter_options : a dictionary with sorter name (key) and
                     a dictionary of kwargs to pass to the
                     spikeinterface sorter class.
    """
    config_dir = Path(os.path.dirname(os.path.realpath(__file__)))

    available_files = glob.glob((config_dir / "*.yaml").as_posix())
    available_files = [Path(path_).stem for path_ in available_files]

    if name not in available_files:  # then assume it is a full path
        assert Path(name).is_file(), (
            f"{name} is neither the name of an existing "
            f"config or valid path to configuration file."
        )

        assert Path(name).suffix in [
            ".yaml",
            ".yml",
        ], f"{name} is not the path to a .yaml file"

        config_filepath = Path(name)

    else:
        config_filepath = config_dir / f"{name}.yaml"

    with open(config_filepath) as file:
        config = yaml.full_load(file)

    pp_steps = config["preprocessing"] if "preprocessing" in config else {}
    sorter_options = config["sorting"] if "sorting" in config else {}
    waveform_options = config["waveforms"] if "waveforms" in config else {}

    utils.cast_pp_steps_values(pp_steps, "tuple")

    return pp_steps, sorter_options, waveform_options



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

import jax.numpy as jnp
import numpy as np
from jax import jit, vmap
from jax.lax import scan

from ..utils import utils

if TYPE_CHECKING:
    from numpy.typing import NDArray
    from spikeinterface import WaveformExtractor


def get_waveform_similarity(
    waveforms: WaveformExtractor, unit_id: int, backend: Literal["numpy", "jax"] = "jax"
):
    """
    Calculate a num_waveform x num_waveform matrix containing
    cosine similarity between all waveforms for a unit. The waveforms
    are saved as 3d array (num_waveforms x num_samples x num_channels) by
    SpikeInterface, and waveforms are averaged across the channels
    where the signal is strongest (see `get_times_of_waveform_spikes()`)
    prior to cosine similarity calculation.

    waveforms : WaveformExtractor
        Spikeinterface WaveformExtractor object.

    unit_id : int
        ID of the unit to calculate the similarity matrix for.

    backend : Literal["jax", "numpy"]
        The backend with which to run waveform comparisons. Jax is around 50
        times faster in this case.
    """
    waveforms_data = waveforms.get_waveforms(unit_id=unit_id)

    selected_waveform_peak_times = get_times_of_waveform_spikes(waveforms, unit_id)

    assert (
        waveforms_data.shape[0] == selected_waveform_peak_times.size
    ), "Number of waveform peak times does not match number of extracted waveforms."

    if waveforms_data.shape[0] == 1:
        utils.message_user(
            f"Skipping {unit_id} as one waveform detected for this unit."
        )
        return np.nan, selected_waveform_peak_times

    if backend == "numpy":
        similarity_matrix = calculate_similarity_numpy(waveforms_data)
    elif backend == "jax":
        similarity_matrix = calculate_similarity_jax(waveforms_data)
    else:
        raise ValueError("`backend` must be: 'jax' or 'numpy'")

    return similarity_matrix, selected_waveform_peak_times


def get_times_of_waveform_spikes(waveforms, unit_id):
    """
    Extract the peak times of the waveforms that were randomly sampled.
    SpikeInterface WaveformExtractor randomly samples a set of spikes
    from all spikes in a unit (default 500) to reduce computation time
    and memory footprint. As such, the index of the sampled waveforms
    is used in the list of all spike times to find the sampled spike times.

    TODO
    ----
    Check manually that the correct spike times are extracted, as compared with Phy.
    """
    select_waveform_tuples = waveforms.get_sampled_indices(unit_id)
    select_waveform_idxs, seg_idxs = zip(*select_waveform_tuples)
    assert np.all(np.array(seg_idxs) == 0), "Multi-segment waveforms not tested."

    all_waveform_peak_idxs = waveforms.sorting.get_unit_spike_train(unit_id)
    selected_waveform_idxs = all_waveform_peak_idxs[np.array(select_waveform_idxs)]
    selected_waveform_peak_times = selected_waveform_idxs / waveforms.sampling_frequency

    return selected_waveform_peak_times


# --------------------------------------------------------------------------------------
# Similarity Matrix Calculators
# --------------------------------------------------------------------------------------


@jit
def calculate_similarity_jax(waveforms_data):
    """
    Speed up calculation of similarity matrix using Jax. First, average
    the waveforms across the best channels (where the waveform signal
    is strongest) for the unit. Then, define a vmapped function that
    calculates the cosine similarity between a waveform and every
    other waveform. Finally, use scan to replicate a for-loop
    over every waveform.

    Note that this redundantly calculates the upper and lower traingular
    of the similarity matrix, but this is necessary to take advantage of vmap
    speed-ups.

    Parameters
    ----------
    waveforms_data : Jax.Array
        A num_waveforms x num_samples x num_channels array containing
        data of all waveforms.

    Returns
    -------
    similarity_matrix : Jax.Array
        num_waveform x num_waveform matrix of cosine similarities
        for the unit.
    """
    averaged_waveforms = jnp.mean(waveforms_data, axis=2)

    def func(carry, i):
        y = averaged_waveforms[i, :]

        sim_matrix_row = vmap(
            lambda x: jnp.dot(x, y.T) / (jnp.linalg.norm(x) * jnp.linalg.norm(y)),
            in_axes=0,
        )(averaged_waveforms)

        return carry, sim_matrix_row

    similarity_matrix = scan(
        func, averaged_waveforms, np.arange(averaged_waveforms.shape[0])
    )[1]

    return similarity_matrix


def calculate_similarity_numpy(waveforms_data: NDArray):
    """
    The similarity matrix is calculated by filling the
    upper triangular only then copying to the lower triangular.

    Parameters
    ----------

    waveforms_data : NDArray
    A num_waveforms x num_samples x num_channels array containing
    data of all waveforms.

    Returns
    -------

    similarity_matrix : NDArray
    num_waveform x num_waveform matrix of cosine similarities
    for the unit.
    """
    num_spikes = waveforms_data.shape[0]
    averaged_waveforms = np.mean(waveforms_data, axis=2)

    similarity_matrix = np.zeros((num_spikes, num_spikes))
    for j in range(num_spikes):
        for i in range(j + 1):
            x = averaged_waveforms[i, :]
            y = averaged_waveforms[j, :]

            similarity_matrix[i, j] = np.dot(x, y.T) / (
                np.linalg.norm(x) * np.linalg.norm(y)
            )

    i_lower = np.tril_indices(similarity_matrix.shape[0], -1)
    similarity_matrix[i_lower] = similarity_matrix.T[i_lower]

    return similarity_matrix



import datetime
import subprocess
from pathlib import Path
from typing import Callable, Dict, Literal, Union

import submitit

from . import utils


def run_job(kwargs, command_func: Callable, command_name: str) -> None:
    """
    Run a job (e.g. run_full_pipeline, run_sorting) on SLURM.

    Parameters
    ----------
    kwargs : Dict
        Keyword arguments passed to run_full_pipeline.

    command_func : Callable
        The function to run (e.g. `run_full_pipeline()`.

    command_name : str
        The name of the command, typically command_func.__name__,
        formatted for logging.
    """
    slurm_opts = kwargs.pop("slurm_batch")
    executor = get_executor(kwargs)

    job = executor.submit(
        wrap_function_with_env_setup, command_func, slurm_opts, **kwargs
    )

    if isinstance(slurm_opts, dict) and "wait" in slurm_opts and slurm_opts["wait"]:
        job.wait()

    send_user_start_message(command_name, job, kwargs)


def run_full_pipeline_slurm(**kwargs) -> None:
    """
    Run the entire preprocessing pipeline in a SLURM job.

    This takes the kwargs passed to the original call of the
    preprocessing function, and feeds them back to the function
    from within the SLURM job.

    Parameters
    ----------
    kwargs : Dict
        Keyword arguments passed to run_full_pipeline.

    Notes
    -----
    The import must occur here to avoid recursive imports.
    """
    from ..pipeline.full_pipeline import run_full_pipeline

    run_job(kwargs, run_full_pipeline, "Full pipeline")


def run_sorting_slurm(**kwargs) -> None:
    """
    Run the sorting pipeline from within a SLURM job.

    See run_full_pipeline_slurm for details, this is identical
    except it is for run_sorting rather than run_full_pipeline.

    Notes
    -----
    The import must occur here to avoid recursive imports.
    """
    from ..pipeline.sort import run_sorting

    run_job(kwargs, run_sorting, "Sorting")


# Utils --------------------------------------------------------------------------------


def get_executor(kwargs: Dict) -> submitit.AutoExecutor:
    """
    Return the executor object that defines parameters
    of the SLURM node to request and the path to
    logs.

    Parameters
    ----------
    kwargs : Dict
        Keyword arguments passed to the main running function
        (e.g. run_full_pipeline, run_sorting).

    Returns
    -------
    executor : submitit.AutoExecutor
        submitit executor object defining requested SLURM
        node parameters.
    """
    log_path = make_job_log_output_path(kwargs)

    print(f"\nThe SLURM batch output logs will " f"be saved to {log_path}\n")

    executor = submitit.AutoExecutor(
        folder=log_path,
    )

    executor.update_parameters(
        nodes=1,
        mem_gb=40,
        timeout_min=24 * 60,
        cpus_per_task=8,
        tasks_per_node=1,
        gpus_per_node=1,
        slurm_gres="gpu:1",
        slurm_partition="gpu",
    )

    return executor


def wrap_function_with_env_setup(
    function: Callable, slurm_opts: Union[Literal[True], Dict], **kwargs
) -> None:
    """
    Set up the environment from within the SLURM job, prior
    to running the processing function (e.g. run_full_pipeline).

    This is required to set up the conda environment within the job
    or the processing function will fail.

    Parameters
    ----------
    function : Callable
        The ephys processing function to run in the SLURM job
        e.g. run_full_pipeline, run_sorting

    slurm_opts : Union[Literal[True], Dict]
        A kwarg passed to the processing function (e.g. run_full_pipeline)
        indicating whether to run in the SLURM job. If True or a Dict,
        the SLURM job is run. If a dict, the environment setup
        can be passed in the 'env_name' field.

    kwargs : Dict
        Keyword arguments passed to the main running function
        (e.g. run_full_pipeline, run_sorting)
    """
    if isinstance(slurm_opts, dict) and "env_name" in slurm_opts:
        env_name = slurm_opts["env_name"]
    else:
        env_name = "swc_ephys"

    print(f"\nrunning {function.__name__} with SLURM....\n")

    subprocess.run(
        f"module load miniconda; " f"source activate {env_name};" f"module load cuda",
        executable="/bin/bash",
        shell=True,
    )

    function(**kwargs)


def make_job_log_output_path(kwargs: Dict) -> Path:
    """
    The SLURM job logs are saved to a folder 'slurm_logs' in the
    base directory in which the processing is being run
    (i.e. the folder containing rawdata, derivatives). .

    Parameters
    ----------
    kwargs : Dict
        Keyword arguments passed to the main running function
        (e.g. run_full_pipeline, run_sorting)

    Returns
    -------
    log_path : Path
        The path to the SLURM log output folder for the current job.
        The logs are saved to a folder with machine datetime.
    """
    now = datetime.datetime.now()

    log_subpath = Path("slurm_logs") / f"{now.strftime('%Y-%m-%d_%H-%M-%S')}"

    if "base_path" in kwargs:
        log_path = kwargs["base_path"] / log_subpath
    else:
        log_path = kwargs["data"].base_path / log_subpath

    log_path.mkdir(exist_ok=True, parents=True)

    return log_path


def send_user_start_message(
    processing_function: str, job: submitit.Job, kwargs: Dict
) -> None:
    """
    Convenience function to print important information
    regarding the SLURM job.

    Parameters
    ----------
    processing_function : str
        The function being run (i.e. run_full_pipeline, run_sorting)

    job : submitit.job
        submitit.job object holding the SLURM job_id

    kwargs : Dict
        Keyword arguments passed to the main running function
        (e.g. run_full_pipeline, run_sorting)
    """
    utils.message_user(
        f"{processing_function} submitted to SLURM with job id {job.job_id}\n"
        f"with arguments{kwargs}"
    )


def is_slurm_installed():
    slurm_installed = subprocess.run("sinfo -v", shell=True).returncode == 0
    return slurm_installed



"""
TODO: this is currently out of date. Do not use.
"""
import argparse

from .pipeline.full_pipeline import run_full_pipeline

# figure out how to give nice asserts on SI asserts that
# are less informative e.g. overwriting sorter output.

parser = argparse.ArgumentParser()

parser.add_argument(
    "base_path",
    type=str,
    help="The base path where the data is stored, containing the rawdata folder.",
)

parser.add_argument(
    "sub_name",
    type=str,
    help="subject name, should match a subject folder name in rawdata.",
)

parser.add_argument(
    "run_name",
    type=str,
    help="spikeglx run name for the subject. Should not contain the gate. e.g.\n"
    "1110925_test_shank1",
)

parser.add_argument(
    "--config-name",
    "--config_name",
    type=str,
    default="default",
    help="name of the configuration file to use.",
)

parser.add_argument(
    "--sorter",
    type=str,
    default="kilosort2_5",
    help="name of the sorter to use.",
)

parser.add_argument(
    "--use-existing-preprocessed-file",
    "--use_existing_preprocessed_file",
    action="store_true",
    help="If true, during sorting an existing preprocessing data folder\n"
    "for the subject will be used if it already exists.",
)


def main():
    args = parser.parse_args()

    run_full_pipeline(
        base_path=args.base_path,
        sub_name=args.sub_name,
        run_name=args.run_name,
        config_name=args.config_name,
        sorter=args.sorter,
        use_existing_preprocessed_file=args.use_existing_preprocessed_file,
    )


if __name__ == "__main__":
    main()



from __future__ import annotations

import copy
import gc
import shutil
import time
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Literal, Tuple, Union

if TYPE_CHECKING:
    from ..data_classes.preprocessing import PreprocessingData
    from ..data_classes.sorting import SortingData

from ..configs.configs import get_configs
from ..pipeline.load_data import load_data_for_sorting
from ..utils import slurm, utils
from ..utils.custom_types import HandleExisting
from .load_data import load_spikeglx_data
from .postprocess import run_postprocess
from .preprocess import preprocess
from .sort import run_sorting


def run_full_pipeline(
    base_path: Union[Path, str],
    sub_name: str,
    run_names: Union[List[str], str],
    config_name: str = "default",
    sorter: str = "kilosort2_5",
    existing_preprocessed_data: HandleExisting = "load_if_exists",
    existing_sorting_output: HandleExisting = "load_if_exists",
    overwrite_postprocessing: bool = False,
    postprocessing_to_run: Union[Literal["all"], Dict] = "all",
    delete_intermediate_files: Tuple[
        Literal["preprocessed", "recording.dat", "temp_wh.dat", "waveforms"]
    ] = (
        "recording.dat",
    ),  # TODO: use new spikeinterface settings for kilosort.
    verbose: bool = True,
    slurm_batch: bool = False,
) -> None:
    """
    Run preprocessing, sorting and post-processing on SpikeGLX data.
    see README.md for detailed information on use. If waveforms and
    postprocessing exist for the subjects / runs, it will be
    overwritten.

    This function must be run in main as uses multiprocessing e.g.
    if __name__ == "__main__":
        run_full_pipeline(args...)

    Parameters
    ----------
    base_path : Union[Path, str]
        Path to the rawdata folder containing subjects folders.

    sub_name : str
        Subject to preprocess. The subject top level dir should reside in
        base_path/rawdata/ .

    run_names : Union[List[str], str],
        The SpikeGLX run name (i.e. not including the gate index). This can
        also be a list of run names. Preprocessing will still occur per-run.
        Runs are concatenated in the order passed prior to sorting.

    config_name : str
        The name of the configuration to use. Note this must be the name
        of a .yaml file (not including the extension) stored in
        swc_ephys/configs.

    sorter : str
        name of the sorter to use e.g. "kilosort2_5".

    existing_preprocessed_data : Literal["overwrite", "load_if_exists", "fail_if_exists"]
        Determines how existing preprocessed data (e.g. from a prior pipeline run)
        is treated.
            "overwrite" : will overwrite any existing preprocessed data output. This will
                          delete the 'preprocessed' folder. Therefore, never save
                          derivative work there.
            "load_if_exists" : will search for existing data and load if it exists.
                               Otherwise, will use the preprocessing from the
                               current run.
            "fail_if_exists" : If existing preprocessed data is found, an error
                               will be raised.

    existing_sorting_output : bool
        Determines how existing sorted data is treated. The same behaviour
        as `existing_preprocessed_data` but for sorting output. If overwrite,
        the 'sorting' folder will be deleted. Therefore, never save
        derivative work there.

    overwrite_postprocessing : bool
        If `False`, an error will be raised if postprocessing output already
        exists. Otherwise, 'postprocessing' folder will be overwritten. Note,
        that the entire 'postprocessing' folder (including all contents) will be
        deleted. Therefore, never save derivative work there.

    postprocessing_to_run : Union[Literal["all"], Dict]
        Specify the postprocessing to run. By default, "all" will run
        all available postprocessing. Otherwise, provide a dict of
        including postprocessing to run e.g. {"quality_metrics: True"}.

    delete_intermediate_files : Tuple[Union["preprocessing", "recording.dat", "temp_wh.dat", "waveforms"]]  # TODO: check types
        Specify intermediate files or folders to delete. This option is useful for
        reducing the size of output data by deleting unneeded files.

        preprocessing  - the 'preprocesed' folder holding the data that has been
                         preprocessed by SpikeInterface
        recording.dat - SpikeInterfaces copies the preprocessed data to folder
                        prior to sorting, where it resides in the 'sorter_output'
                        folder. Often, this can be deleted after sorting.
        temp_wh.dat - Kilosort output file that holds the data preprocessed by
                      Kilosort (e.g. drift correction). By default, this is used
                      for visualisation in Phy.
        waveforms - The waveform outputs that SpikeInterface generates to calculate
                    quality metrics. Often, these can be deleted once final quality
                    metrics are computed.

    verbose : bool
        If True, messages will be printed to console updating on the
        progress of preprocessing / sorting.

    slurm_batch : bool
        If True, the pipeline will be run in a SLURM job. Set False
        if running on an interactive job, or locally.
    """
    if slurm_batch:
        local_args = copy.deepcopy(locals())
        slurm.run_full_pipeline_slurm(**local_args)
        return
    assert slurm_batch is False, "SLURM run has slurm_batch set True"

    pp_steps, sorter_options, waveform_options = get_configs(config_name)

    preprocess_data = load_spikeglx_data(base_path, sub_name, run_names)

    preprocess_data = preprocess(preprocess_data, pp_steps, verbose)

    save_preprocessed_data_if_required(preprocess_data, existing_preprocessed_data)

    sorting_data = run_or_get_sorting(
        preprocess_data, existing_sorting_output, sorter, sorter_options, verbose
    )

    sorting_data.set_sorter_output_paths(sorter)
    delete_postprocessing_output_if_it_exists(sorting_data, overwrite_postprocessing)

    run_postprocess(
        sorting_data,
        sorter,
        existing_waveform_data="fail_if_exists",
        postprocessing_to_run=postprocessing_to_run,
        verbose=verbose,
        waveform_options=waveform_options,
    )

    handle_delete_intermediate_files(sorting_data, delete_intermediate_files)


def handle_delete_intermediate_files(sorting_data, delete_intermediate_files):
    """ """
    if "preprocessed" in delete_intermediate_files:
        # remove the existing link to the preprocessed data binary.
        # wait time of 5 s is arbitrary.
        # TODO: this feels very hacky. Can unlink the memory map from the segment?
        # Ask SI.
        del sorting_data["0-preprocessed"]
        gc.collect()
        time.sleep(5)

        if sorting_data.preprocessed_data_path.is_dir():
            shutil.rmtree(sorting_data.preprocessed_data_path)

    if "recording.dat" in delete_intermediate_files:
        if (
            recording_file := sorting_data.sorter_run_output_path / "recording.dat"
        ).is_file():
            recording_file.unlink()

    if "temp_wh.dat" in delete_intermediate_files:
        if (
            recording_file := sorting_data.sorter_run_output_path / "temp_wh.dat"
        ).is_file():
            recording_file.unlink()

    if "waveforms" in delete_intermediate_files:
        if (
            waveforms_path := sorting_data.postprocessing_output_path / "waveforms"
        ).is_dir():
            shutil.rmtree(waveforms_path)


def delete_postprocessing_output_if_it_exists(
    sorting_data: SortingData, overwrite_postprocessing: bool
):
    """
    If previous postprocessing output exists, it must be deleted before
    the new postprocessing is run. As a safety measure, `overwrite_postprocessing`
    must be set to `True` to perform the deletion.
    """
    if sorting_data.postprocessing_output_path.is_dir():
        if overwrite_postprocessing:
            utils.message_user(
                f"Deleting existing postprocessing "
                f"output at {sorting_data.postprocessing_output_path}"
            )
            shutil.rmtree(sorting_data.postprocessing_output_path)
        else:
            raise RuntimeError(
                f"Postprocessing output already exists at "
                f"{sorting_data.postprocessing_output_path} "
                f"but `overwrite_postprocessing` is `False`. Setting "
                f"`overwrite_postprocessing` will delete the postprocessing "
                f"folder and all it's contents."
            )


def run_or_get_sorting(
    preprocess_data: PreprocessingData,
    existing_sorting_output: HandleExisting,
    sorter: str,
    sorter_options: Dict,
    verbose: bool,
) -> SortingData:
    """
    Handle existing sorting output. If previous output exists, load, error or
    overwrite according to `existing_sorting_output`. See `run_full_pipeline()` for details.
    """
    # TODO: "sorting" to configs. In general handle this function better
    # it should not be on preprocess_data. It could be a classmethod on
    # SortingData but this may be even messier.
    sorting_path = preprocess_data.get_expected_sorter_path(sorter) / "sorting"

    if sorting_path.is_dir() and existing_sorting_output == "load_if_exists":
        utils.message_user(f"Loaded pre-existing sorting output from {sorting_path}")

        sorting_data = load_data_for_sorting(
            preprocess_data.preprocessed_data_path,
        )

    elif sorting_path.is_dir() and existing_sorting_output == "fail_if_exists":
        raise RuntimeError(
            f"Sorting output already exists at {sorting_path} and"
            f"`existing_sorting_output` is set to 'fail_if_exists'."
        )

    else:
        utils.message_user("Running sorting...")

        overwrite_existing_sorter_output = True

        sorting_data = run_sorting(
            preprocess_data.preprocessed_data_path,
            sorter,
            sorter_options,
            overwrite_existing_sorter_output,
            verbose,
        )

    return sorting_data


def save_preprocessed_data_if_required(
    preprocess_data: PreprocessingData,
    existing_preprocessed_data: Literal[
        "overwrite", "load_if_exists", "fail_if_exists"
    ],
) -> None:
    """
    Handle the loading of existing preprocessed data.
    See `run_full_pipeline()` for details.
    """
    preprocess_path = preprocess_data.preprocessed_data_path

    if existing_preprocessed_data == "overwrite":
        if preprocess_path.is_dir():
            utils.message_user(f"Removing existing file at {preprocess_path}\n")

        utils.message_user(f"Saving preprocessed data to {preprocess_path}")

        preprocess_data.save_all_preprocessed_data(overwrite=True)

    elif existing_preprocessed_data == "load_if_exists":
        if preprocess_path.is_dir():
            utils.message_user(
                f"\nSkipping preprocessing, using file at "
                f"{preprocess_path} for sorting.\n"
            )
        else:
            utils.message_user(
                f"No data found at {preprocess_path}, saving" f"preprocessed data."
            )
            preprocess_data.save_all_preprocessed_data(overwrite=False)

    elif existing_preprocessed_data == "fail_if_exists":
        if preprocess_path.is_dir():
            raise FileExistsError(
                f"Preprocessed binary already exists at "
                f"{preprocess_path}. "
                f"To overwrite, set 'existing_preprocessed_data' to 'overwrite'"
            )
        preprocess_data.save_all_preprocessed_data(overwrite=False)

    else:
        raise ValueError(
            "`existing_preproessed_data` argument not recognised."
            "Must be: 'load_if_exists', 'fail_if_exists' or 'overwrite'."
        )



from pathlib import Path

from swc_ephys.pipeline.load_data import load_data_for_sorting
from swc_ephys.pipeline.visualise import visualise

preprocessing_path = r"X:\neuroinformatics\scratch\jziminski\ephys\test_data\steve_multi_run\1119617\time-short\derivatives\1119617\1119617_LSE1_shank12_posttest1_pretest1\preprocessed"

sorting_data = load_data_for_sorting(
    Path(preprocessing_path),
)
visualise(sorting_data, time_range=(1, 2))



from __future__ import annotations

import copy
import os
import shutil
from pathlib import Path
from typing import TYPE_CHECKING, Dict, Literal, Optional, Tuple, Union

if TYPE_CHECKING:
    from ..data_classes.sorting import SortingData

import platform

import spikeinterface.sorters as ss

from ..pipeline.load_data import load_data_for_sorting
from ..utils import slurm, utils


def run_sorting(
    preprocessed_data_path: Union[Path, str],
    sorter: str = "kilosort2_5",
    sorter_options: Optional[Dict] = None,
    overwrite_existing_sorter_output: bool = False,
    verbose: bool = True,
    slurm_batch: bool = False,
) -> SortingData:
    """
    Run a sorter on pre-processed data. Takes a PreprocessingData (pipeline.data_class)
    object that contains spikeinterface recording objects for the preprocessing
    pipeline (or path to existing 'preprocessed' output folder.

    Here, save the preprocessed recording to binary file. Then, run sorting
    on the saved binary. The preprocessed binary and sorting output are
    saved in a 'derivatives' folder, in the same top-level folder as 'rawdata'.
    The folder structure will be the same as in 'rawdata'.

    Parameters
    ----------
    preprocessed_data_path : Union[Path, str]
        Path to previously saved 'preprocessed' directory.

    sorter : str
        Name of the sorter to use (e.g. "kilosort2_5").

    sorter_options : Dict
        Kwargs to pass to spikeinterface sorter class.

    overwrite_existing_sorter_output : bool
         If False, an error will be raised if sorting output already
         exists. If True, existing sorting output will be overwritten.

    verbose : bool
        If True, messages will be printed to consolve updating on the
        progress of preprocessing / sorting.

    slurm_batch : bool
        If True, the pipeline will be run in a SLURM job. Set False
        if running on an interactive job, or locally.

    """
    preprocessed_data_path = Path(preprocessed_data_path)

    sorting_data = load_data_for_sorting(preprocessed_data_path)
    sorting_data.set_sorter_output_paths(sorter)

    if slurm_batch:
        local_args = copy.deepcopy(locals())
        slurm.run_sorting_slurm(**local_args)
        return sorting_data

    sorter_options_dict = validate_inputs(slurm_batch, sorter, sorter_options, verbose)

    # Load preprocessed data from saved preprocess output path.
    utils.message_user(
        f"\nLoading binary preprocessed data from {preprocessed_data_path.as_posix()}\n"
    )

    # This must be run from the folder that has both sorter output AND rawdata
    os.chdir(sorting_data.base_path)

    singularity_image, docker_image = get_image_run_settings(sorter)

    utils.message_user(f"Starting {sorter} sorting...")

    # TODO: can remove on API change PR #1908
    # Also, once #1908 is merged and updated, we need to check
    # that 'delete_intermediate_files' is not passed by the
    # user in the config file because it is overridden here.
    if "kilosort" in sorter:
        sorter_options_dict.update({"delete_tmp_files": False})

    ss.run_sorter(
        sorter,
        sorting_data.data["0-preprocessed"],
        output_folder=sorting_data.sorting_output_path,
        singularity_image=singularity_image,
        docker_image=docker_image,
        remove_existing_folder=overwrite_existing_sorter_output,
        **sorter_options_dict,
    )

    move_singularity_image_if_required(sorting_data, singularity_image, sorter)

    return sorting_data


def move_singularity_image_if_required(
    sorting_data: SortingData,
    singularity_image: Optional[Union[Literal[True], str]],
    sorter: str,
) -> None:
    """
    On Linux, images are cased to the sorting_data base folder
    by default by SpikeInterface. To avoid re-downloading
    images, these are moved to a pre-determined folder (home
    for local, pre-set on an HPC). This is only required
    for singularity, as docker-desktop handles all image
    storage.

    Parameters
    ----------

    singularity_image: Optional[Union[Literal[True], Path]]
        Holds either a path to an existing (stored) sorter, or
        `True`. If `True`, no stored sorter image exists and so
        we move it. The next time sorting is performed, it will use
        this stored image.

    sorter : str
        Name of the sorter.
    """
    if singularity_image is True:
        assert platform.system() != "Windows", "Docker should be used on windows."
        store_singularity_image(sorting_data.base_path, sorter)


def get_image_run_settings(
    sorter: str,
) -> Tuple[
    Optional[Union[Literal[True], str]], Optional[bool]
]:  # cannot set this to Literal[True], for unknown reason.
    """
    Determine how to run the sorting, either locally or in a container
    if required (e.g. kilosort2_5). On windows, Docker is used,
    otherwise singularity. Docker images are handled by Docker-desktop,
    but singularity image storage is handled internally, see
    `move_singularity_image_if_required()`.

    Parameters
    ----------

    sorter : str
        Sorter name.
    """
    can_run_locally = ["spykingcircus", "mountainsort5", "tridesclous"]

    if sorter in can_run_locally:
        singularity_image = docker_image = None
    else:
        if platform.system() == "Windows":
            singularity_image = None
            docker_image = True
        else:
            singularity_image = get_singularity_image(sorter)
            docker_image = None

    return singularity_image, docker_image


def store_singularity_image(base_path: Path, sorter: str) -> None:
    """
    When running locally, SpikeInterface will pull the docker image
    to the current working directly. Move this to home/.swc_ephys
    so they can be used again in future and are centralised.

    Parameters
    ----------
    base_path : Path
        Base-path on the SortingData object, the path that holds
        `rawdata` and `derivatives` folders.

    sorter : str
        Name of the sorter for which to store the image.
    """
    path_to_image = base_path / utils.get_sorter_image_name(sorter)
    shutil.move(path_to_image, utils.get_local_sorter_path(sorter).parent)


def get_singularity_image(sorter: str) -> Union[Literal[True], str]:
    """
    Get the path to a pre-installed system singularity image. If none
    can be found, set to True. In this case SpikeInterface will
    pull the image to the current working directory, and
    this will be moved after sorting
    (see store_singularity_image).

    Parameters
    ----------
    sorter : str
        Name of the sorter to get the image for.

    Returns
    -------
    singularity_image [ Union[Literal[True], str]
        If `str`, the path to the singularity image. Otherwise if `True`,
        this tells SpikeInterface to pull the image.
    """
    singularity_image: Union[Literal[True], str]

    if utils.get_hpc_sorter_path(sorter).is_file():
        singularity_image = str(utils.get_hpc_sorter_path(sorter))

    elif utils.get_local_sorter_path(sorter).is_file():
        singularity_image = str(utils.get_local_sorter_path(sorter))
    else:
        singularity_image = True

    return singularity_image


def validate_inputs(
    slurm_batch: bool, sorter: str, sorter_options: Optional[Dict], verbose: bool
) -> Dict:
    """
    Check that the sorter is valid, singularity is installed and format
    the dictionary of options to pass to the sorter.

    Parameters
    ----------
    slurm_batch : bool
        Whether the run is a SLURM batch. This must be False, otherwise
        indicates some recursion has occurred when SLURM run itself
        calls this function.

    sorter : str
        Name of the sorter.

    sorter_options : Optional[Dict]
        Options to pass to the SpikeInterface sorter. If None, no options
        are passed.

    verbose : bool
        Whether SpikeInterface sorting is run in 'verbose' mode.

    Returns
    -------
    sorter_options_dict : Dict
        A dictionary of configurations to run the sorter with, these
        are passed to SpikeInterface sorter.
    """
    assert slurm_batch is False, "SLURM run has slurm_batch set True"

    supported_sorters = [
        "kilosort2",
        "kilosort2_5",
        "kilosort3",
        "mountainsort5",
        "spykingcircus",
        "tridesclous",
    ]

    assert (
        sorter in supported_sorters
    ), f"sorter {sorter} is invalid, must be one of: {supported_sorters}"

    assert (
        utils.check_singularity_install()
    ), "Singularity must be installed to run sorting."

    sorter_options_dict = {}
    if sorter_options is not None and sorter in sorter_options:
        sorter_options_dict = sorter_options[sorter]

    sorter_options_dict.update({"verbose": verbose})

    return sorter_options_dict



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Dict, List, Literal, Tuple, Union

if TYPE_CHECKING:
    from spikeinterface.core import BaseRecording

    from ..data_classes.preprocessing import PreprocessingData
    from ..data_classes.sorting import SortingData

import copy
import os.path
import subprocess
from pathlib import Path

import numpy as np
from spikeinterface import concatenate_recordings

# TODO: how does splitting by group work for sorting?


def canonical_names(name: str) -> str:
    """
    Store the canonical names e.g. filenames, tags
    that are used throughout the project. This setup
    means filenames can be edited without requiring
    extensive code changes.

    Parameters
    ----------
    name : str
        short-hand name of the full name of interest.

    Returns
    -------
    filenames[name] : str
        The full name of interest e.g. filename.

    """
    filenames = {
        "preprocessed_yaml": "preprocess_data_attributes.yaml",
    }
    return filenames[name]


def get_keys_first_char(
    data: Union[PreprocessingData, SortingData], as_int: bool = False
) -> Union[List[str], List[int]]:
    """
    Get the first character of all keys in a dictionary. Expected
    that the first characters are integers (as str type).

    Parameters
    ----------
    data : Union[PreprocessingData, SortingData]
        swc_ephys PreprocessingData class holding filepath information.

    as_int : bool
        If True, the first character of the keys are cast to
        integer type.

    Returns
    -------
    list_of_numbers : Union[List[str], List[int]]
        A list of numbers of string or integer type, that are
        the first numbers of the Preprocessing / Sorting Data
        .data dictionary keys.
    """
    list_of_numbers = [
        int(key.split("-")[0]) if as_int else key.split("-")[0] for key in data.keys()
    ]
    return list_of_numbers


def get_dict_value_from_step_num(
    data: Union[PreprocessingData, SortingData], step_num: str
) -> Tuple[BaseRecording, str]:
    """
    Get the value of the PreprocessingData dict from the preprocessing step number.

    PreprocessingData contain keys indicating the preprocessing steps,
    starting with the preprocessing step number.
    e.g. 0-raw, 1-raw-bandpass_filter, 2-raw_bandpass_filter-common_average

    Return the value of the dict (spikeinterface recording object)
    from the dict using only the step number.

    Parameters
    ----------
    data : Union[PreprocessingData, SortingData]
        swc_ephys PreprocessingData class holding filepath information.

    step_num : str
        The preprocessing step number to get the value (i.e. recording object)
        from.

    Returns
    -------
    dict_value : BaseRecording
        The SpikeInterface recording stored in the dict at the
        given preprocessing step number.

    pp_key : str
        The key of the preprocessing dict at the given
        step number.
    """
    if step_num == "last":
        pp_key_nums = get_keys_first_char(data, as_int=True)

        # Complete overkill as a check but this is critical.
        step_num = str(int(np.max(pp_key_nums)))
        assert (
            int(step_num) == len(data.keys()) - 1
        ), "the last key has been taken incorrectly"

    select_step_pp_key = [key for key in data.keys() if key.split("-")[0] == step_num]

    assert len(select_step_pp_key) == 1, "pp_key must always have unique first char"

    pp_key: str = select_step_pp_key[0]
    dict_value = data[pp_key]

    return dict_value, pp_key


def message_user(message: str, verbose: bool = True) -> None:
    """
    Method to interact with user.

    Parameters
    ----------
    message : str
        Message to print.

    verbose : bool
        The mode of the application. If verbose is False,
        nothing is printed.
    """
    if verbose:
        print(message)


def concatenate_runs(recording: BaseRecording) -> BaseRecording:
    """
    Convenience function to concatenate the segments
    of a recording object.

    Parameters
    ----------
    recording : BaseRecording
        A spikeinterface recording object.

    Returns
    -------
    concatenated_recording : BaseRecording
        The SpikeInterface recording object with all
        segments concatenated into a single segments.
    """
    message_user(f"Concatenating {recording.get_num_segments()} into a single segment.")

    concatenated_recording = concatenate_recordings([recording])

    return concatenated_recording


def get_local_sorter_path(sorter: str) -> Path:
    """
    Return the path to a sorter singularity image. The sorters are
    stored by swc_ephys in the home folder.

    Parameters
    ----------
    sorter : str
        The name of the sorter to get the path to (e.g. kilosort2_5).

    Returns
    ---------
    local_path : Path
        The path to the sorter image on the local machine.
    """
    local_path = (
        Path.home() / ".swc_ephys" / "sorter_images" / get_sorter_image_name(sorter)
    )
    local_path.parent.mkdir(exist_ok=True, parents=True)
    return local_path


def get_hpc_sorter_path(sorter: str) -> Path:
    """
    Return the path to the sorter image on the SWC HCP (ceph).

    Parameters
    ----------
    sorter : str
        The name of the sorter to get the path to (e.g. kilosort2_5).

    Returns
    -------
    sorter_path : Path
        The base to the sorter image on SWC HCP (ceph).
    """
    base_path = Path("/ceph/neuroinformatics/neuroinformatics/scratch/sorter_images")
    sorter_path = base_path / sorter / get_sorter_image_name(sorter)
    return sorter_path


def get_sorter_image_name(sorter: str) -> str:
    """
    Get the sorter image name, as defined by how
    SpikeInterface names the docker images it provides.

    Parameters
    ----------
    sorter : str
        The name of the sorter to get the path to (e.g. kilosort2_5).

    Returns
    -------
    sorter_name : str
        The SpikeInterface filename of the docker image for that sorter.
    """
    if "kilosort" in sorter:
        sorter_name = f"{sorter}-compiled-base.sif"
    else:
        if sorter == "spykingcircus":
            sorter = "spyking-circus"
        sorter_name = f"{sorter}-base.sif"
    return sorter_name


def check_singularity_install() -> bool:
    """
    Check the system install of singularity.

    Returns
    -------
    is_installed : bool
        Indicates whether Singularity is installed on the machine.
    """
    try:
        subprocess.run("singularity --version", shell=True)
        is_installed = True
    except FileNotFoundError:
        is_installed = False

    return is_installed


def sort_list_of_paths_by_datetime_order(list_of_paths: List[Path]) -> List[Path]:
    """
    Given a list of paths to folders, sort the paths by the creation
    time of the folders they point to. Return the sorted
    list of paths.

    Parameters
    ----------
    list_of_paths : List[Path]
        A list of paths to sort into datetime order.


    Returns
    -------
    list_of_paths_by_creation_time : List[Path]
        A list containing `list_of_paths` ordered by the
        folder creation timestamp.
    """
    list_of_paths_by_creation_time = copy.deepcopy(list_of_paths)
    list_of_paths_by_creation_time.sort(key=os.path.getctime)

    return list_of_paths_by_creation_time


def list_of_files_are_in_datetime_order(
    list_of_paths: List[Path], creation_or_modification: str = "creation"
) -> bool:
    """
    Assert whether a list of paths are in order. By default, check they are
    in order by creation date. Can also check if they are ordered by
    modification date.

    Parameters
    ----------
    list_of_paths : List[Path]
        A list of paths to sort into datetime order.

    creation_or_modification : str
        If "creation", check the list of paths are ordered by creation datetime.
        Otherwise if "modification", check they are sorterd by modification datetime.

    Returns
    -------
    is_in_time_order : bool
        Indicates whether `list_of_paths` was in creation or modification time order.
        depending on the value of `creation_or_modification`.
    """
    assert creation_or_modification in [
        "creation",
        "modification",
    ], "creation_or_modification must be 'creation' or 'modification."

    filter: Callable
    filter = (
        os.path.getctime if creation_or_modification == "creation" else os.path.getmtime
    )

    list_of_paths_by_time = copy.deepcopy(list_of_paths)
    list_of_paths_by_time.sort(key=filter)

    is_in_time_order = list_of_paths == list_of_paths_by_time

    return is_in_time_order


def make_sorter_base_output_path(base_path, sub_name, pp_run_name, sorter):
    """
    Make the canonical sorter output path.
    """
    sorter_base_output_path = (
        base_path / "derivatives" / sub_name / f"{pp_run_name}" / f"{sorter}-sorting"
    )
    return sorter_base_output_path


def make_preprocessing_plot_title(
    run_name: str,
    full_key: str,
    shank_idx: int,
    recording_to_plot: BaseRecording,
    total_used_shanks: int,
) -> str:
    """
    For visualising data, make the plot titles (with headers in bold). If
    more than one shank is used, the title will also contain information
    on the displayed shank.

    Parameters
    ----------
    run_name : str
        The name of the preprocessing run (e.g. "1-phase_shift").

    full_key : str
        The full preprocessing key (as defined in preprocess.py).

    shank_idx : int
        The SpikeInterface group number representing the shank number.

    recording_to_plot : BaseRecording
        The SpikeInterface recording object that is being displayed.

    total_used_shanks : int
        The total number of shanks used in the recording. For a 4-shank probe,
        this could be between 1 - 4 if not all shanks are mapped.

    Returns
    -------
    plot_title : str
        The formatted plot title.
    """
    plot_title = (
        r"$\bf{Run \ name:}$" + f"{run_name}"
        "\n" + r"$\bf{Preprocessing \ step:}$" + f"{full_key}"
    )
    if total_used_shanks > 1:
        plot_title += (
            "\n"
            + r"$\bf{Shank \ group:}$"
            + f"{shank_idx}, "
            + r"$\bf{Num \ channels:}$"
            + f"{recording_to_plot.get_num_channels()}"
        )
    return plot_title


def cast_pp_steps_values(
    pp_steps: Dict, list_or_tuple: Literal["list", "tuple"]
) -> None:
    """
    The settings in the pp_steps dictionary that defines the options
    for preprocessing should be stored in Tuple as they are not to
    be edited. However, when dumping Tuple to .yaml, there are tags
    displayed on the .yaml file which are very ugly.

    These are not shown when storing list, so this function serves
    to convert Tuple and List values in the preprocessing dict when
    loading / saving the preprocessing dict to .yaml files. This
    function converts `pp_steps` in place.

    Parameters
    ----------
    pp_steps : Dict
        The dictionary indicating the preprocessing steps to perform.

    list_or_tuple : Literal["list", "tuple"]
        The direction to convert (i.e. if "tuple", will convert to Tuple).
    """
    assert list_or_tuple in ["list", "tuple"], "Must cast to `list` or `tuple`."
    func = tuple if list_or_tuple == "tuple" else list

    for key in pp_steps.keys():
        pp_steps[key] = func(pp_steps[key])


# Misc. --------------------------------------------------------------------------------


def get_probe_num_groups(data: Union[PreprocessingData, SortingData]) -> int:
    """
    Get the number of probe groups on a recording-objects `probe` attribute.
    This is typically the number of shanks on the probe, which are treated
    separately by SI.

    By default, uses the first step on the recording, as probe metadata will not
    change throughout preprocessing.

    Parameters
    ----------
    data : Union[PreprocessingData, SortingData]
        The data object on which the recordings are stored.

    Returns
    -------
    num_groups : int
        The number of groups on the probe associated with the data recordings.
    """
    num_groups = np.unique(data[data.init_data_key].get_property("group")).size
    return num_groups



from pathlib import Path

from swc_ephys.pipeline.sort import run_sorting

base_path = Path(
    r"/ceph/neuroinformatics/neuroinformatics/scratch/jziminski/ephys/test_data/steve_multi_run/1119617/time-mid"
)
sub_name = "1119617"
run_name = "1119617_LSE1_shank12_posttest1_pretest1"

preprocessed_data_path = (
    base_path / "derivatives" / sub_name / f"{run_name}" / "preprocessed"
)


if __name__ == "__main__":
    # sorting uses multiprocessing so must be in __main__
    run_sorting(
        preprocessed_data_path,
        sorter="kilosort2_5",
        sorter_options={"kilosort2_5": {"car": False}},
        overwrite_existing_sorter_output=False,
        slurm_batch=False,
    )



"""
TODO: these tests don't check any output, only that things run without error
"""
import os
import shutil
from pathlib import Path

import pytest
from spikeinterface import concatenate_recordings

from swc_ephys.pipeline import full_pipeline, preprocess
from swc_ephys.pipeline.full_pipeline import get_configs
from swc_ephys.pipeline.load_data import load_spikeglx_data
from swc_ephys.utils.slurm import is_slurm_installed

CAN_SLURM = is_slurm_installed()


class TestFirstEphys:
    @pytest.fixture(scope="class")
    def output_data_path(self):
        script_path = Path(os.path.dirname(os.path.realpath(__file__)))
        data_path = script_path.parent
        output_data_path = data_path / "data" / "steve_multi_run"
        return output_data_path

    @pytest.fixture(scope="function")
    def test_info(self, output_data_path, request):
        """ """
        if not hasattr(request, "param"):
            mode = "time-short"
        else:
            mode = request.param

        output_data_path = output_data_path / mode

        sub_name = "1119617"
        run_names = [
            "1119617_LSE1_shank12",
            "1119617_posttest1_shank12",
            "1119617_pretest1_shank12",
        ]

        output_path = output_data_path / "derivatives"
        if output_path.is_dir():
            shutil.rmtree(output_path)

        yield [output_data_path, sub_name, run_names, output_path]

        if output_path.is_dir():
            shutil.rmtree(output_path)

    def run_full_pipeline(
        self,
        base_path,
        sub_name,
        run_names,
        existing_preprocessed_data="fail_if_exists",
        existing_sorting_output="fail_if_exists",
        slurm_batch=False,
        sorter="kilosort2_5",
    ):
        full_pipeline.run_full_pipeline(
            base_path,
            sub_name,
            run_names,
            config_name="default",
            sorter=sorter,
            existing_preprocessed_data=existing_preprocessed_data,
            existing_sorting_output=existing_sorting_output,
            overwrite_postprocessing=True,
            slurm_batch=slurm_batch,
        )

    @pytest.mark.parametrize("test_info", ["time-tiny"], indirect=True)
    def test_preprocessing_options_with_small_file(self, test_info):
        """"""
        pp_steps, __, __ = get_configs("test_pp_small_file")

        preprocess_data = load_spikeglx_data(*test_info[:3])

        preprocess_data = preprocess.preprocess(preprocess_data, pp_steps, verbose=True)
        preprocess_data.save_all_preprocessed_data(overwrite=True)

    def test_preprocessing_options_with_large_file(self, test_info):
        """
        Some preprocessing steps do not ru non the  short file because
        of issues with chunk size. The ones that didn't work
        are run here on a larger file.
        """
        pp_steps, __, __ = get_configs("test_pp_large_file")

        preprocess_data = load_spikeglx_data(*test_info[:3])
        # motion correction requires only 1 segment
        preprocess_data["0-raw"] = concatenate_recordings([preprocess_data["0-raw"]])

        preprocess_data = preprocess.preprocess(preprocess_data, pp_steps, verbose=True)
        preprocess_data.save_all_preprocessed_data(overwrite=True)

    @pytest.mark.parametrize(
        "sorter",
        [
            "kilosort2",
            "kilosort2_5",
            "kilosort3",
            "mountainsort5",
            "spykingcircus",
            "tridesclous",
        ],
    )
    def test_single_run_local__(self, test_info, sorter):
        test_info.pop(3)
        test_info[2] = test_info[2][0]
        self.run_full_pipeline(*test_info, sorter=sorter)

    def test_single_run_local_overwrite(self, test_info):
        test_info.pop(3)
        test_info[2] = test_info[2][0]

        self.run_full_pipeline(*test_info)

        self.run_full_pipeline(*test_info, existing_preprocessed_data="overwrite")

        with pytest.raises(BaseException) as e:
            self.run_full_pipeline(
                *test_info, existing_preprocessed_data="fail_if_exists"
            )

        assert "To overwrite, set 'existing_preprocessed_data' to 'overwrite'" in str(
            e.value
        )

    def test_multi_run_local(self, test_info):
        test_info.pop(3)

        test_info[2] = test_info[2][0]

        self.run_full_pipeline(*test_info)

    @pytest.mark.skipif(CAN_SLURM is False, reason="CAN_SLURM is false")
    def test_single_run_slurm(self, test_info, output_data_path):
        test_info.pop(3)

        test_info[2] = test_info[2][0]

        self.clear_slurm_logs(output_data_path)

        self.run_full_pipeline(*test_info, slurm_batch={"wait": True})

        self.check_slurm_log(output_data_path)

    def check_slurm_log(self, output_data_path):
        slurm_run = output_data_path.glob("slurm_logs/*/*log.out")
        slurm_run = list(slurm_run)[0]

        with open(slurm_run, "r") as log:
            log_output = log.readlines()

        log_output = " ".join(log_output)

        assert "Stopping container" in log_output
        assert "Saving waveforms to" in log_output
        assert "Quality metrics saved to" in log_output
        assert "Job completed successfully" in log_output

    @pytest.mark.skipif(CAN_SLURM is False, reason="CAN_SLURM is false")
    def test_multi_run_slurm(self, test_info, output_data_path):
        test_info.pop(3)

        self.clear_slurm_logs(output_data_path)

        self.run_full_pipeline(*test_info, slurm_batch={"wait": True})

        self.check_slurm_log(output_data_path)

    def clear_slurm_logs(self, output_data_path):
        slurm_path = output_data_path / "slurm_logs"
        [shutil.rmtree(path_) for path_ in slurm_path.glob("*-*-*_*-*-*")]

    def test_preprocessing_exists_error(self):
        raise NotImplementedError

    def test_use_existing_preprocessing_errror(self):
        raise NotImplementedError

    def test_sorter_exists_error(self):
        raise NotImplementedError

    def test_overwrite_sorter(self):
        raise NotImplementedError

    def test_sorting_only_local(self):
        raise NotImplementedError

    def test_sorting_only_slumr(self):
        raise NotImplementedError



import fnmatch
import os
import shutil
import warnings
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import yaml

from ..utils import utils
from .base import BaseUserDict


class PreprocessingData(BaseUserDict):
    def __init__(
        self,
        base_path: Union[Path, str],
        sub_name: str,
        run_names: Union[List[str], str],
    ):
        """
        Dictionary to store SpikeInterface preprocessing recordings.
        These are lazy and preprocessing only run when the recording.get_traces()
        is called, or the data is saved to binary.

        Details on the preprocessing steps are held in the dictionary keys e.g.
        e.g. 0-raw, 1-raw-bandpass_filter, 2-raw_bandpass_filter-common_average
        and recording objects are held in the value.

        The class manages paths to raw data and preprocessing output,
        as defines methods to dump key information and the SpikeInterface
        binary to disk.

        Parameters
        ----------
        base_path : Union[Path, str]
            Path where the rawdata folder containing subjects.

        sub_name : str
            'subject' to preprocess. The subject top level dir should
            reside in base_path/rawdata/.

        run_names : Union[List[str], str]
            The SpikeGLX run name (i.e. not including the gate index)
            or list of run names.
        """
        super(PreprocessingData, self).__init__()

        self.top_level_folder = "rawdata"
        self.init_data_key = "0-raw"

        self.base_path, checked_run_names, self.rawdata_path = self.validate_inputs(
            base_path,
            sub_name,
            run_names,
        )

        self.sub_name = sub_name

        (
            self.all_run_paths,
            self.all_run_names,
            self.pp_run_name,
        ) = self.create_runs_from_single_or_multiple_run_names(checked_run_names)

        self.pp_steps: Optional[Dict] = None
        self.data: Dict = {"0-raw": None}
        self.sync = None

        self.preprocessed_data_path: Path
        self._pp_data_attributes_path: Path
        self._pp_binary_data_path: Path
        self._sync_channel_data_path: Path
        self._set_preprocessing_output_path()

    # Handle Multiple Runs -------------------------------------------------------------

    def create_runs_from_single_or_multiple_run_names(
        self,
        run_names: List[str],
    ) -> Tuple[List[Path], List[str], str]:
        """
        The run_names may be a single run, or a list of runs to process.
        Return a list of paths to the runs found on the system. enforces
        that the runs are in datetime order.

        Parameters
        ----------
        run_names : Union[List[str], str]
            The SpikeGLX run name (i.e. not including the gate index) or
            list of run names.

        Returns
        -------
        all_run_paths : List[Path]
            List of full filepaths for all runs used in the pipeline.

        all_run_names : List[str]
            List of run names (folder names within the subject level folder)
            to be used in the pipeline.

        run_name : str
            The name of the run used in the derivatives. For a single run,
            this will be the name of the run. When run_names is a list of run
            names, this will be an amalgamation of all run names
            (see self.make_run_name_from_multiple_run_names)
        """
        if len(run_names) > 1:
            all_run_paths, run_name = self.get_multi_run_names_paths(run_names)
        else:
            run_name = run_names[0]
            all_run_paths = [self.get_sub_folder_path() / f"{run_name}_g0"]

        assert len(run_names) > 0, (
            f"No runs found for {run_names}. "
            f"Make sure to specify run_names without gate / trigger (e.g. no _g0)."
        )

        all_run_names = [path_.stem for path_ in all_run_paths]

        for run_path in all_run_paths:
            assert run_path.is_dir(), (
                f"The run folder {run_path.stem} cannot be found at "
                f"file path {run_path.parent}"
            )

        utils.message_user(f"The order of the loaded runs is:" f"{all_run_names}")

        return all_run_paths, all_run_names, run_name

    def get_multi_run_names_paths(
        self,
        run_names: List[str],
    ) -> Tuple[List[Path], str]:
        """
        Get the paths to the runs when there is more than one run specified.
        If it is a list of run names, they are searched, checked exist and
        assert in datetime order.

        Parameters
        ----------
        run_names : Union[List[str], str]
            The spikeglx run name (i.e. not including the gate index) or
            list of run names.

        Returns
        -------
        all_run_paths : List[Path]
            List of full filepaths for all runs used in the pipeline.

        all_run_names : List[str]
            List of run names (folder names within the subject level folder)
            to be used in the pipeline.

        run_name : str
            The name of the run used in the derivatives. For a single run,
            this will be the name of the run. When run_names is a list of run names,
            this will be an amalgamation of all run names
            (see self.make_run_name_from_multiple_run_names)
        """
        all_run_paths = [
            self.get_sub_folder_path() / f"{name}_g0" for name in run_names
        ]

        if not utils.list_of_files_are_in_datetime_order(all_run_paths, "creation"):
            warnings.warn(
                "The runs provided are not in creation datetime order.\n"
                "They will be concatenated in the order provided."
            )

        pp_run_name = self.make_run_name_from_multiple_run_names(run_names)

        return all_run_paths, pp_run_name

    def make_run_name_from_multiple_run_names(self, run_names: List[str]) -> str:
        """
        Make a single run_name given a list of run names. This will use the
        first part of the first name and then add unique parts of the
        subsequent names to the string.

        Parameters
        ----------
        run_names : Union[List[str], str]
            A list of run names.

        Returns
        -------
        pp_run_name : str
            A single run name formed from the list of run names.
        """
        all_names = []
        for idx, name in enumerate(run_names):
            if idx == 0:
                all_names.extend(name.split("_"))
            else:
                split_name = name.split("_")
                new_name = [n for n in split_name if n not in all_names]
                all_names.extend(new_name)

        if "g0" in all_names:
            all_names.remove("g0")

        pp_run_name = "_".join(all_names)

        return pp_run_name

    def get_expected_sorter_path(self, sorter: str) -> Path:
        return utils.make_sorter_base_output_path(
            self.base_path, self.sub_name, self.pp_run_name, sorter
        )

    # Load and Save --------------------------------------------------------------------

    def set_pp_steps(self, pp_steps: Dict) -> None:
        """
        Set the preprocessing steps (`pp_steps`) attribute
        that defines the preprocessing steps and options.

        Parameters
        ----------
        pp_steps : Dict
            Preprocessing steps to setp as class attribute. These are used
            when `pipeline.preprocess.preprocess()` function is called.
        """
        self.pp_steps = pp_steps

    def save_all_preprocessed_data(self, overwrite: bool = False) -> None:
        """
        Save the preprocessed output data to binary, as well
        as important class attributes to a .yaml file.

        Both are saved in a folder called 'preprocessed'
        in derivatives/<sub_name>/<pp_run_name>

        Parameters
        ----------
        overwrite : bool
            If `True`, existing preprocessed output will be overwritten.
            By default, SpikeInterface will error if a binary recording file
            (`si_recording`) already exists where it is trying to write one.
            In this case, an error  should be raised before this function
            is called.
        """
        if overwrite:
            if self.preprocessed_data_path.is_dir():
                shutil.rmtree(self.preprocessed_data_path)
        self._save_data_class()
        self._save_preprocessed_binary()
        self._save_sync_channel()

    def _save_preprocessed_binary(self) -> None:
        """
        Save the fully preprocessed data (i.e. last step in the
        preprocessing chain) to binary file. This is required for sorting.
        """
        recording, __ = utils.get_dict_value_from_step_num(self, "last")
        recording.save(folder=self._pp_binary_data_path, chunk_memory="10M")

    def _save_sync_channel(self) -> None:
        """ """
        assert self.sync is not None, "Sync channel on PreprocessData is None"
        self.sync.save(folder=self._sync_channel_data_path, chunk_memory="10M")

    def _save_data_class(self) -> None:
        """
        Save the key attributes of this class to a .yaml file.
        """
        assert self.pp_steps is not None, "type narrow `pp_steps`."

        utils.cast_pp_steps_values(self.pp_steps, "list")

        attributes_to_save = {
            "base_path": self.base_path.as_posix(),
            "sub_name": self.sub_name,
            "pp_run_name": self.pp_run_name,
            "pp_steps": self.pp_steps,
            "all_run_paths": [path_.as_posix() for path_ in self.all_run_paths],
            "all_run_names": [name for name in self.all_run_names],
            "preprocessed_data_path": self.preprocessed_data_path.as_posix(),
            "_pp_binary_data_path": self._pp_binary_data_path.as_posix(),
            "_pp_data_attributes_path": self._pp_data_attributes_path.as_posix(),
            "_sync_channel_data_path": self._sync_channel_data_path.as_posix(),
        }
        if not self.preprocessed_data_path.is_dir():
            os.makedirs(self.preprocessed_data_path)

        with open(
            self._pp_data_attributes_path,
            "w",
        ) as attributes:
            yaml.dump(attributes_to_save, attributes, sort_keys=False)

    # Handle Paths ---------------------------------------------------------------------

    def _set_preprocessing_output_path(self) -> None:
        """
        Set the folder tree where preprocessing output will be
        saved. This is canonical and should not change.

        TODO: move this to a canonical_filepaths module.
        """
        self.preprocessed_data_path = (
            self.base_path
            / "derivatives"
            / self.sub_name
            / f"{self.pp_run_name}"
            / "preprocessed"
        )

        self._pp_data_attributes_path = (
            self.preprocessed_data_path / utils.canonical_names("preprocessed_yaml")
        )
        self._pp_binary_data_path = self.preprocessed_data_path / "si_recording"
        self._sync_channel_data_path = self.preprocessed_data_path / "sync_channel"

    def get_sub_folder_path(self) -> Path:
        """
        Get the path to the rawdata subject folder.

        Returns
        -------
        sub_folder_path : Path
            Path to the self.sub_name folder in the `rawdata` path folder.
        """
        sub_folder_path = Path(self.base_path / self.top_level_folder / self.sub_name)
        return sub_folder_path

    # Validate Inputs ------------------------------------------------------------------

    def validate_inputs(
        self, base_path: Union[str, Path], sub_name: str, run_names: Union[str, list]
    ) -> Tuple[Path, List[str], Path]:
        """
        Check the rawdata path, subject path exists and ensure run_names
        is a list of strings without SpikeGLX gate number of the run names.

        Parameters
        ----------
        base_path : Union[str, Path]
            Path to the base folder in which `rawdata` folder containing
            all rawdata (i.e. list of subject names) are held.

        sub_name : str
            'subject' name to preprocess data for.

        run_names : List[str]
            List of run names to process, in order they should be
            processed / concatenated.

        Returns
        -------
        base_path : Path
            `base_path` definitely as a Path object.

        run_names : List[str]
            Validated `run_names` as a List.

        rawdata_path : Path
            Path to the canonical `rawdata` path required for
            future processing.
        """
        base_path = Path(base_path)
        rawdata_path = base_path / self.top_level_folder

        assert (rawdata_path / sub_name).is_dir(), (
            f"Subject directory not found. {sub_name} "
            f"is not a folder in {rawdata_path}"
        )

        assert rawdata_path.is_dir(), (
            f"Ensure there is a folder in base path called 'rawdata'.\n"
            f"No rawdata directory found at {rawdata_path}\n"
            f"where subject-level folders are placed."
        )
        if not isinstance(run_names, list):
            run_names = [run_names]

        for name in run_names:
            assert "g0" not in name.split("_"), (
                f"gate index should not be on the run name. Remove _g0 from\n" f"{name}"
            )

        for name in run_names:
            gate_idx_in_name = fnmatch.filter(name.split("_"), "g?")
            assert len(gate_idx_in_name) == 0, (
                f"Gate with index larger than 0 is not supported. This is found "
                f"in run name {name}. "
            )
        return base_path, run_names, rawdata_path



from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional, Tuple, Union

if TYPE_CHECKING:
    import matplotlib
    from numpy.typing import NDArray
    from spikeinterface.core import BaseRecording


import matplotlib.pyplot as plt
import numpy as np
import spikeinterface.widgets as sw

from ..data_classes.preprocessing import PreprocessingData
from ..data_classes.sorting import SortingData
from ..utils import utils


def visualise(
    data: Union[PreprocessingData, SortingData],
    steps: Union[List[str], str] = "all",
    mode: str = "auto",
    as_subplot: bool = False,
    time_range: Optional[Tuple] = None,
    show_channel_ids: bool = False,
    run_number: int = 1,
) -> None:
    """
    Plot the data at various preprocessing steps, useful for quality-checking.
    Takes the pipeline.data_class.PreprocessingData object
    (output from pipeline.preprocess).

    Channels are displayed ordered by depth. Note preprocessing is lazy, and only the
    section if data displayed will be preprocessed.

    If multiple preprocessing steps are shown, they will be placed in subplots
    on a single plot if as_subplots is True, otherwise in separate plots.

    data : swc_ephys PreprocessingData class containing the preprocessing output (a dict
           of keys indicating the preprocessing step and values are spikeinterface
           recording objects.

    steps : the preprocessing steps to show, specified as a number (e.g. "1"),
            or list of numbers ["1", "2", "3"] or "all" to show multiple steps.

    mode : "line" to show line plot, "map" to show as image, or "auto" to determine
            based on the number of displayed channels.

    as_subplot : if True, multiple preprocessing steps will be displayed as subplots
                 on the same plot. Otherwise, they will be plot as separate plots.

    time_range : time range of data to display in seconds e.g. (1, 5) will display the
                 range 1- 5 s.

    show_channel_ids : if True, channel IDS will be displayed on the plot.

    run_number : The run number to visualise (in the case of a concatenated recording.
                 Under the hood, each run maps to a SpikeInterface segment_index.
    """
    steps, as_subplot = process_input_arguments(data, steps, as_subplot)

    total_used_shanks = utils.get_probe_num_groups(data)

    for shank_idx in range(total_used_shanks):
        if as_subplot:
            fig, ax, num_rows, num_cols = generate_subplot(steps)

        for idx, step in enumerate(steps):
            recording, full_key = utils.get_dict_value_from_step_num(data, str(step))

            validate_options_against_recording(recording, data, time_range, run_number)

            recordings = recording.split_by(property="group")
            recording_to_plot = recordings[shank_idx]

            run_name = get_run_name(data, run_number)

            plot_title = utils.make_preprocessing_plot_title(
                run_name,
                full_key,
                shank_idx,
                recording_to_plot,
                total_used_shanks,
            )

            current_ax = (
                None if not as_subplot else get_subplot_ax(idx, ax, num_rows, num_cols)
            )

            sw.plot_timeseries(
                recording_to_plot,
                order_channel_by_depth=False,
                time_range=time_range,
                return_scaled=True,
                show_channel_ids=show_channel_ids,
                mode=mode,
                ax=current_ax,
                segment_index=run_number - 1,
            )

            if current_ax is None:
                plt.title(plot_title)
                plt.show()
            else:
                current_ax.set_title(plot_title)

        if as_subplot:
            plt.show()


def generate_subplot(
    steps: Union[List[str], str]
) -> Tuple[matplotlib.figure.Figure, NDArray[matplotlib.axes._axes.Axes], int, int]:
    """
    Generate the Matplotlib subplots figure, with hte number of
    columns fixed at two and the number of rows depending on the
    number of steps.

    Parameters
    ----------
    steps : Union[List[str], str]
        the preprocessing steps that will be plot.

    Returns
    -------
    fig : matplotlib.figure.Figure
        Matplotlib figure displaying the preprocessed data.

    ax :  NDArray[matplotlib.axes._axes.Axes]
        An array holding the axes objects in a plot / subplot.

    num_rows : int
    num_cols : int
        Number of rows and columns in the subplot, as determined
        by the number of processing steps to display.
    """
    num_cols = 2
    num_rows = np.ceil(len(steps) / num_cols).astype(int)
    fig, ax = plt.subplots(num_rows, num_cols)
    fig.tight_layout(rect=[0, 0.03, 1, 0.95])

    return fig, ax, num_rows, num_cols


def get_subplot_ax(
    idx: int, ax: NDArray[matplotlib.axes._axes.Axes], num_rows: int, num_cols: int
) -> matplotlib.axes._axes.Axes:
    """
    Get the Axes object of the current axis to plot. As preprocessing
    steps are cycled through, axis are plot separately based on the
    current row / column needs to plot. We need to convert
    from subscript to linear indices to get the appropriate axis object.

    Parameters
    ----------
    idx : int
        Idx of the currently displayed preprocessing step.

    ax : NDArray[matplotlib.axes._axes.Axes]
        An array of axis objects on the plot / subplot.

    num_rows : int
    num_cols : int
        The number of rows and columns in the subplot (as determined by the
        number of preprocessing steps).

    Returns
    -------
    current_ax : matplotlib.axes._axes.Axes
        Axis to plot the current preprocessing step data on.
    """
    idx_unraveled = np.unravel_index(idx, shape=(num_rows, num_cols))
    current_ax = ax[idx_unraveled]
    return current_ax


def process_input_arguments(
    data: Union[PreprocessingData, SortingData],
    steps: Union[List[str], str],
    as_subplot: bool,
) -> Tuple[Union[List[str], str], bool]:
    """
    Check the passed configurations are valid.
    See `visualise()` for arguments.

    Returns
    -------
    steps : Dict
        `steps` as a List, checked for validity.

    as_subplot: bool
        `as_subplot`, possibly forced to False if the number
        of steps is only 1 (in which case subplot is redundant).
    """
    if not isinstance(steps, List):
        steps = [steps]

    if "all" in steps:
        assert len(steps) == 1, "if using 'all' only put one step input"
        steps = utils.get_keys_first_char(data)  # type: ignore

    assert len(steps) <= len(data), (
        "The number of steps must be less or equal to the "
        "number of steps in the recording"
    )

    if len(steps) == 1 and as_subplot:
        as_subplot = False

    return steps, as_subplot


def validate_options_against_recording(
    recording: BaseRecording,
    data: Union[PreprocessingData, SortingData],
    time_range: Optional[Tuple],
    run_number: int,
) -> None:
    """
    Check the passed configurations are valid.
    See `visualise()` for arguments.

    TODO
    ----
    can't find a better way to get final timepoint,
    but must be somewhere, this is wasteful.
    """
    if isinstance(data, PreprocessingData):
        num_runs = len(data.all_run_names)
        assert run_number <= num_runs, (
            "The run_number must be less than or equal to the "
            "number of runs specified."
        )
    if time_range is not None:
        assert (
            time_range[1] <= recording.get_times(segment_index=run_number - 1)[-1]
        ), "The time range specified is longer than the maximum time of the recording."


def get_run_name(data: Union[PreprocessingData, SortingData], run_number: int) -> str:
    """
    Get the name of the run of which the data is being displayed.
    If the object is PreprocessingData, the data is not concatenated
    and it is one of the rawdata runs for the subject.

    Otherwise it is a SortingData object, of which there is only one, concatenated
    output run.

    Parameters
    ----------
    data : Union[PreprocessingData, SortingData]
        Preprocessing or sorting data dictionary.

    run_number : int
        Run of the run to plot.

    Returns
    -------
    run_name : str
        Name of the run. For PreprocessingData, this will be the raw
        data run name. For SortingData, if multiple runs are
        concatenated it will be an amalgamation of the rawdata
        runs used to generated it.
    """
    if isinstance(data, PreprocessingData):
        run_name = data.all_run_names[run_number - 1]
    elif isinstance(data, SortingData):
        run_name = data.pp_run_name

    return run_name



import warnings
from pathlib import Path
from typing import Dict, Union

import spikeinterface as si
import yaml

from ..utils import utils
from .base import BaseUserDict


class SortingData(BaseUserDict):
    """
    Class to organise the sorting of preprocessed data. Handles the
    paths to preprocessed data, sorted output and all post-processing
    steps.

    This class should be agnostic to any rawdata and the PreprocessingData
    class, and concerned only with preprocessed data onwards.

    Parameters
    ----------
    preprocessed_data_path : Union[str, Path]
        Path to the preprocessed data output path, that contains
        the dumped si_recording and .yaml file of PreprocessingData
        attributes.

    Notes
    -----
    This class matches the data-access signature of PreprocessData
    for compatibility with `visualise()` methods. This pattern
    does not really make sense for this class as it most likely will
    only ever contain a single data attribute.This includes
    some jiggery-pokery with the `self.init_data_key`. This is not a nice
    pattern and can be improved on refactoring of `visualise.py`.
    """

    def __init__(self, preprocessed_data_path: Union[str, Path]):
        """ """
        super(SortingData, self).__init__()

        self.preprocessed_data_path = Path(preprocessed_data_path)
        self.check_preprocessed_data_path_exists()

        self.top_level_folder = "derivatives"
        self.init_data_key = "0-preprocessed"
        self.pp_info = self.load_preprocess_data_attributes()

        self.base_path = Path(self.pp_info["base_path"])
        self.show_warning_if_base_path_diverged()

        self.sub_name = self.pp_info["sub_name"]
        self.pp_run_name = self.pp_info["pp_run_name"]

        # These paths are set when the sorter
        # is known, set_sorter_output_paths()
        self.sorter_base_output_path: Path
        self.sorter_run_output_path: Path
        self.postprocessing_output_path: Path
        self.quality_metrics_path: Path
        self.unit_locations_path: Path

        # This is set later, depending on
        # concatenated or not.
        self.data: Dict = {"0-preprocessed": None}

    def show_warning_if_base_path_diverged(self):
        """
        It is expected that the passed preprocessed data output path
        is in the same location as the data was saved during preprocessed,
        as stored in the PreprocessData attribute. This can be broken
        however, in the case of accessing the same folder as a mounted drive.
        """
        pp_base_path = [
            path
            for path in self.preprocessed_data_path.parents
            if path.stem == "derivatives"
        ][0]
        if pp_base_path != self.base_path:
            warnings.warn(
                f"The base path of the `preprocessed_data_path` does not match the "
                f"`base_path` contained used to run the preprocessing. This is expected "
                f"if running the same folder from a different location (e.g. mounted drive). "
                f"Otherwise, check the base paths are correct.\n"
                f"passed base path: {self.preprocessed_data_path}\n"
                f"original base path: {self.base_path}"
            )

    def check_preprocessed_data_path_exists(self) -> None:
        """
        Ensure the preprocessed data path exists, otherwise
        it cannot be loaded and sorting will fail.
        """
        if not self.preprocessed_data_path.is_dir():
            raise FileNotFoundError(
                f"No preprocessed data found at " f"{self.preprocessed_data_path}"
            )

    def load_preprocess_data_attributes(self) -> Dict:
        """
        Load the PreprocessingData attributes that were
        saved when writing the preprocessed data to file.

        These fields are used to set the sorting and other
        output paths, and for provenance.
        """
        with open(
            Path(self.preprocessed_data_path)
            / utils.canonical_names("preprocessed_yaml")
        ) as file:
            pp_info = yaml.full_load(file)
        return pp_info

    def load_preprocessed_binary(self, concatenate: bool = True):
        """
        Use SpikeInterface to load the binary-data into a recording object.

        Parameters
        ----------
        concatenate : bool
            If `True`, all segments in the loaded SpikeInterface
            recording object will be concatenated into a single
            segment.
        """
        binary_path = self.preprocessed_data_path / "si_recording"  # TODO: configs

        if not binary_path.is_dir():
            raise FileNotFoundError(
                f"No preprocessed SI binary-containing folder "
                f"found at {binary_path}."
            )
        recording = si.load_extractor(binary_path)

        if concatenate:
            recording = utils.concatenate_runs(recording)

        self.data["0-preprocessed"] = recording

    def set_sorter_output_paths(self, sorter: str) -> None:
        """
        Set the sorter-specific output paths. The same data may be
        sorted multiple times by different sorters.

        sort_base_output_path : str
            canonical name, is where spikeinterface
            automatically saves sorter output
        """
        self.sorter_base_output_path = utils.make_sorter_base_output_path(
            self.base_path, self.sub_name, self.pp_run_name, sorter
        )

        self.sorting_output_path = self.sorter_base_output_path / "sorting"
        self.sorter_run_output_path = self.sorting_output_path / "sorter_output"
        self.postprocessing_output_path = (
            self.sorter_base_output_path / "postprocessing"
        )
        self.quality_metrics_path = (
            self.postprocessing_output_path / "quality_metrics.csv"
        )
        self.unit_locations_path = (
            self.postprocessing_output_path / "unit_locations.csv"
        )



from pathlib import Path

from swc_ephys.pipeline.postprocess import run_postprocess

base_path = Path(
    r"/ceph/neuroinformatics/neuroinformatics/scratch/jziminski/ephys/test_data/steve_multi_run/1119617/time-short"
)
# TODO: why is this taking the preprocessed path? isn't sorting more intuitive? or top level? think...
sub_name = "1119617"
run_name = "1119617_LSE1_shank12_posttest1_pretest1"

preprocessing_path = (
    base_path / "derivatives" / sub_name / f"{run_name}" / "preprocessed"
)

run_postprocess(
    preprocessing_path,
    sorter="kilosort2_5",
    existing_waveform_data="overwrite",
    postprocessing_to_run="all",
)
